"""
    Memory Asset
"""

from starfish.asset.asset_base import AssetBase

class MemoryAsset(AssetBase):
    """

    Memory asset can be used to try out assets without accessing any file system

    :param data: data string or byte text to save as the asset
    :type data: str or byte array
    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str

    """
    def __init__(self, metadata=None, did=None, data=None):
        default_metadata = {
            'name': 'MemoryAsset',
            'type': 'data',
            'contentType': 'application/octet-stream',
        }
        if metadata is None:
            metadata = default_metadata
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')
        metadata = AssetBase.merge_metadata(metadata, default_metadata)

        if data:
            if isinstance(data, str):
                metadata['contentType'] = 'text/plain; charset=utf-8'
            else:
                metadata['contentType'] = 'application/octet-stream'
        AssetBase.__init__(self, 'data', metadata, did)
        self._data = data

    def save(self, filename):
        """
        Saves the data in the memory asset to a file.

        :param str filename: Filename to save the data.

        """

        if self._data:
            with open(filename, 'wb') as fp:
                fp.write(self._data)

    @property
    def data(self):
        """

        Return the asset data

        :return: the asset data
        :type: str or byte
        """
        return self._data



"""
    Memory Asset
"""
import os
from mimetypes import MimeTypes

from starfish.asset.asset_base import AssetBase

class FileAsset(AssetBase):
    """

    File asset can be used manage a file asset on the ocean network

    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str
    :param str filename: filename of the asset to register

    """
    def __init__(self, metadata=None, did=None, filename=None):
        default_metadata = {
            'name': 'FileAsset',
            'type': 'file',
            'author': 'File Asset',
            'contentType': 'application/octet-stream',
        }
        if metadata is None:
            metadata = default_metadata
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')
        metadata = AssetBase.merge_metadata(metadata, default_metadata)

        AssetBase.__init__(self, 'file', metadata, did)
        if filename:
            self._filename = filename
            if not os.path.exists(filename):
                raise NotFoundErr(f'File {filename} not found')
            mime = MimeTypes()
            mime_type = mime.guess_type(f'file://{self._filename}')
            if mime_type:
                self._metadata['contentType'] = mime_type[0]
            self._metadata['contentLength'] = os.path.getsize(self._filename)
            self._metadata['filename'] = os.path.basename(self._filename)

    @property
    def filename(self):
        """

        Return the asset data

        :return: the asset data
        :type: str or byte
        """
        return self._filename

    @property
    def data(self):
        if os.path.exists(self._filename):
            with open(self._filename, 'rb') as fp:
                return fp.read()



"""
    Inovke Asset
"""

from starfish.asset.asset_base import AssetBase

class OperationAsset(AssetBase):

    """

    Operation asset can be used to perform a invokable service

    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str

    """
    def __init__(self, metadata=None, did=None):
        default_metadata = {
            'name': 'OperationAsset',
            'type': 'operation',
        }
        if metadata is None:
            metadata = default_metadata
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')
        metadata = AssetBase.merge_metadata(metadata, default_metadata)

        AssetBase.__init__(self, 'operation', metadata, did)

        if not self.is_asset_type('operation'):
            raise ValueError('The metadata type is not a valid type for this asset')


    def is_mode(self, mode_type):
        """

        Check to see if this operation supports the mode provided.

        :param str mode_type: Mode type to check to see if this operation supports
        :return: Return True if this mode is supported
        :type: boolean

        """
        try:
            return mode_type in self._metadata['operation']['modes']
        except:
            raise ValueError('Metadata does not contain operation->modes structure')



"""
    test_13_asset_listing_search

    As a developer building client code to interact with an Ocean marketplace,
    I need a way to search available asset listings

"""

import secrets

from starfish.asset import MemoryAsset


def test_13_asset_listing_search(resources, surfer_agent):
    test_data = secrets.token_hex(1024)
    asset = MemoryAsset(data=test_data)
    listing = surfer_agent.register_asset(asset, resources.listing_data)
    assert(listing)
    assert(listing.asset)



"""
    test_05_agent_endpoint_update


    As a developer managing a Ocean Agent,
    I need to be able to update service endpoints for my Agent
 
"""

import re

from starfish.agent import SurferAgent
from starfish.ddo.starfish_ddo import StarfishDDO



def test_05_agent_endpoint_update(ocean, config, surfer_agent):
    new_endpoint_uri = '/app/v99/meta/test'
    endpoint = surfer_agent.get_endpoint('metadata')
    assert(endpoint)
    assert(re.search('meta/data', endpoint))

    ddo = SurferAgent.generate_ddo(config.surfer_url)
    ddo.set_service_endpoint('Ocean.Meta.v1', f'{config.surfer_url}{new_endpoint_uri}')
    new_agent = SurferAgent(ocean, ddo=ddo)
    new_endpoint = new_agent.get_endpoint('metadata')
    assert(new_endpoint)
    assert(re.search(new_endpoint_uri, new_endpoint))



#!/usr/bin/env python
"""

    Tests for `starfish-py account`.

"""
import unittest
import os
import logging
import secrets

from starfish import Ocean, logger
from squid_py import Ocean as SquidOcean
from squid_py import Config as SquidConfig
from squid_py.keeper.web3_provider import Web3Provider


def test_account_load(ocean, config):
    # address only
    account = ocean.get_account(config.publisher_account['address'])
    assert(account)
    assert(account.address == config.publisher_account['address'])
    assert(account.password is None)

    # address and password as a tuple
    account = ocean.get_account(config.publisher_account)
    assert(account)
    assert(account.address == config.publisher_account['address'])
    assert(account.password == config.publisher_account['password'])

def test_account_get_squid_account(ocean, config):
    account = ocean.get_account(config.publisher_account)
    assert(account)
    squid_account = account._squid_account
    assert(squid_account)
    assert(account.is_hosted)

def test_account_get_balance(ocean, config):
    account = ocean.get_account(config.publisher_account)
    assert(account)
    assert(account.unlock())
    account.request_tokens(1)
    assert(account.ocean_balance)
    assert(account.ether_balance)
    print(account.ether_balance)

def test_account_list(ocean, config):
    publisher_account = ocean.get_account(config.publisher_account)
    accounts = ocean.accounts
    assert(len(accounts) > 1)
    assert(publisher_account.is_address_equal((list(accounts)[0])))

def _disable_test_account_creation_and_transfer(ocean, config):
    password = secrets.token_hex(120)
    account = ocean.create_account(password)
    assert(account)
    found_account = ocean.get_account(account.address)

    assert(found_account)

    search_account = None
    for address in ocean.accounts:
        if account.address == address:
            search_account = address
    assert(search_account)
    assert(account.ether_balance == 0)
    assert(account.ocean_balance == 0)

    # test ether transfer
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()
    publisher_account.transfer_ether(account.address, 3)
    assert(account.ether_balance == 3)

    # test ocean token request
    account.unlock()
    account.request_tokens(7)
    assert(account.ocean_balance == 7)

    # test transfer ocean tokens
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()
    publisher_account.transfer_token(account.address, 5)
    assert(account.ocean_balance == 12)






"""
    test_16_asset_bundle

    As a developer working with asset bundles
    I want a way to get to sub-assets.
    
"""

import secrets
import logging
import json

from starfish.asset import MemoryAsset
from starfish.asset import BundleAsset

TEST_ASSET_COUNT = 4


def test_16_asset_bundle_register(ocean, config, surfer_agent, metadata):
    pass
        




"""

Squid Helpers

"""
from squid_py.brizo.brizo_provider import BrizoProvider
from tests.integration.mocks.brizo_mock import BrizoMock


"""
def setup_squid_purchase(ocean, listing, account):
    BrizoProvider.set_brizo_class(BrizoMock)
    mock = BrizoProvider.get_brizo()

    model = ocean.get_squid_model()
    ddo = model._squid_ocean.assets.resolve(listing.asset.did)
    mock.subscribe(ocean, account._squid_account, listing.asset.did, ddo)
"""



"""
    Asset
"""

from starfish.asset.asset_base import AssetBase

class Asset(AssetBase):

    def __init__(self, metadata=None, did=None, filename=None):
        AssetBase.__init__(self, 'asset', metadata, did)





from starfish.asset.asset_base import AssetBase
from starfish.asset.memory_asset import MemoryAsset
from starfish.asset.bundle_asset import BundleAsset
from starfish.asset.operation_asset import OperationAsset
from starfish.asset.file_asset import FileAsset
from starfish.asset.remote_asset import RemoteAsset
from starfish.asset.asset import Asset


def create_asset_from_metadata(metadata, did=None):
    """
    Create a new asset class based on the metadata. Once created assign
    the metadata to the asset and also the optional did

    :param dict metadata: metadata to test and create the correct asset object
    :param str did: optional did to assign to the asset

    :return: Return an asset class based on the metadata type
    """

    asset_type = AssetBase.get_asset_type(metadata)

    if asset_type == 'memory' or asset_type == 'data':
        return MemoryAsset(metadata, did)
    elif asset_type == 'bundle':
        return BundleAsset(metadata, did)
    elif asset_type == 'operation':
        return OperationAsset(metadata, did)
    elif asset_type == 'file':
        return FileAsset(metadata, did)
    elif asset_type == 'remote':
        return RemoteAsset(metadata, did)
    return Asset(metadata, did)



"""
    SurferModel - Model to access the Surfer Services
"""
import io
import json
import requests
from web3 import Web3

from squid_py.did_resolver.did_resolver import DIDResolver

from starfish import logger

# default base URI for this version surfer
SURFER_BASE_URI = '/api/v1'

SUPPORTED_SERVICES = {
    'metadata': {
        'type': 'Ocean.Meta.v1',
        'uri': f'{SURFER_BASE_URI}/meta/data',
    },
    'storage': {
        'type': 'Ocean.Storage.v1',
        'uri': f'{SURFER_BASE_URI}/assets',
    },
    'invoke': {
        'type': 'Ocean.Invoke.v1',
        'uri': f'{SURFER_BASE_URI}/invoke',
    },
    'market': {
        'type': 'Ocean.Market.v1',
        'uri': f'{SURFER_BASE_URI}/market',
    },
    'trust': {
        'type': 'Ocean.Trust.v1',
        'uri': f'{SURFER_BASE_URI}/trust',
    },
    'auth': {
        'type': 'Ocean.Auth.v1',
        'uri': f'{SURFER_BASE_URI}/auth',
    },
}

INVOKE_SYNC_METHOD = '/invoke'
INVOKE_ASYNC_METHOD = '/invokeasync'
INVOKE_JOB_METHOD = '/jobs'

class SurferModel():
    _http_client = requests

    def __init__(self, ocean, did=None, ddo=None, options=None):
        """init a standard ocan connection, with a given DID"""
        self._ocean = ocean
        self._did = did
        self._ddo = ddo
        if not options:
            options = {}

        self._headers = {'content-type': 'application/json'}
        authorization = options.get('authorization')
        if authorization: # this is an OAuth2 token
            self._headers['Authorization'] = f'token {authorization}'

    def register_asset(self, metadata):
        """
        Register an asset with the agent storage server
        :param metadata: metadata to write to the storage server

        :return: A dict of the following items
        [0] asset_id.
        [1] did of the asset.
        [2] metadata in text format, that was saved.

        :type: dict
        """
        result = None
        saved_asset_id = self.save_metadata(metadata)
        result = {
                'asset_id': saved_asset_id,
                'metadata': metadata,
            }
        return result

    def save_metadata(self, metadata):
        """save metadata to the agent server, using the asset_id and metadata"""
        url = self.get_endpoint('metadata')
        logger.debug(f'metadata save url {url}')
        response = SurferModel._http_client.post(url, json=metadata, headers=self._headers)
        if response and response.status_code == requests.codes.ok:

            data = response.json()
            logger.debug(f'metadata asset response returned {data}')
            return data
        else:
            msg = f'metadata asset response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def create_listing(self, asset_id, listing_data):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings'
        data = {
            'assetid': asset_id,
            'info': listing_data,
        }
        response = SurferModel._http_client.post(url, json=data, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = response.json()
            logger.debug('listing response returned: ' + str(data))
            return data
        else:
            msg = f'listing response failed: {response.status_code} {response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def download_asset(self, url):
        response = SurferModel._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = response.content
            return data
        else:
            msg = f'GET assets response failed: {response.status_code} {response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_asset_store_url(self, asset_id):
        endpoint = self.get_endpoint('storage')
        url = f'{endpoint}/{asset_id}'
        return url

    def get_listing(self, listing_id):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings/{listing_id}'
        response = SurferModel._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = response.json()
            return data
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_listings(self):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings'
        response = SurferModel._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = response.json()
            return data
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def update_listing(self, listing_id, data):
        endpoint = self.get_endpoint('market')
        url = f'{endpoint}/listings/{listing_id}'
        response = SurferModel._http_client.put(url, json=data, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            return True
        else:
            msg = f'PUT listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def upload_asset_data(self, url, asset_id, data):
        logger.debug(f'uploading data to {url}')
        data_bytes = data
        if not isinstance(data_bytes, bytes):
            data_bytes = data.encode()

        files = { 'file':  ( asset_id, io.BytesIO(data_bytes), 'application/octet-stream') }
        headers = {
            'Authorization': self._headers['Authorization']
        }
        response = SurferModel._http_client.post(url, files=files, headers=headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == requests.codes.created):
            return True
        else:
            msg = f'upload asset response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def read_metadata(self, asset_id):
        """read the metadata from a service agent using the asset_id"""

        result = None
        endpoint = self.get_endpoint('metadata')
        url = f'{endpoint}/{asset_id}'
        logger.debug(f'metadata read url {url}')
        response = SurferModel._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            result = {
                'asset_id': asset_id,
                'metadata_text': response.content
            }
            # convert to str if bytes
            if isinstance(result['metadata_text'], bytes):
                result['metadata_text'] = response.content.decode('utf-8')
        else:
            logger.warning(f'metadata asset read {asset_id} response returned {response}')
        return result

    def purchase_asset(self, purchase):
        """record purchase"""
        url = self.get_endpoint('market') + '/purchases'
        logger.debug(f'market url for purchases {url}')
        response = SurferModel._http_client.post(url, json=purchase, headers=self._headers)
        if response and response.status_code == requests.codes.ok:

            json = response.json()
            logger.debug(f'purchase response returned {json}')
            return json
        else:
            msg = f'purchase response failed: {response.status_code}'
            logger.error(msg)
            # NOTE surfer may return error information in additon to a 500
            if response:
                json = response.json()
                logger.error(f'purchase response returned {json}')
            raise ValueError(msg)
        return None


    def invoke(self, asset_id, params, is_async=False):
        """

        call the invoke based on the asset_id without leading 0x
        """
        endpoint = self.get_endpoint('invoke')

        if is_async:
            url = f'{endpoint}{INVOKE_ASYNC_METHOD}/{asset_id}'
        else:
            url = f'{endpoint}{INVOKE_SYNC_METHOD}/{asset_id}'
        print(f'request {url}')
        response = SurferModel._http_client.post(url, json=params, headers=self._headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == 201):
            json = response.json()
            logger.debug('invoke response returned: ' + str(json))
            return json
        else:
            msg = f'invoke response failed: {response.status_code} {response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_job(self, job_id):
        endpoint = self.get_endpoint('invoke')
        url = f'{endpoint}{INVOKE_JOB_METHOD}/{job_id}'
        response = SurferModel._http_client.get(url, headers=self._headers)
        if response and response.status_code == requests.codes.ok:
            data = response.json()
            return data
        else:
            msg = f'GET job response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None


    def get_endpoint(self, name):
        """return the endpoint based on the name of the service or service type"""
        supported_service = SurferModel.find_supported_service(name)
        if supported_service is None:
            message = f'unknown surfer endpoint service name or type: {name}'
            logger.error(message)
            raise ValueError(message)

        service_type = supported_service['type']
        endpoint = None
        if self._ddo:
            service = self._ddo.get_service(service_type)
            if not service:
                message = f'unable to find surfer endpoint service type {service_type}'
                logger.error(message)
                raise ValueError(message)
            endpoint = service.endpoints.service
        if not endpoint:
            message = f'unable to find surfer endpoint for {name} = {service_type}'
            logger.error(message)
            raise ValueError(message)
        return endpoint


    @property
    def ddo(self):
        """return the DDO stored for this agent"""
        return self._ddo

    @property
    def did(self):
        """return the DID used for this agent"""
        return self._did

    @property
    def register_name(self):
        return self._register_name

    @staticmethod
    def is_metadata_valid(asset_id, metadata_text):
        """
        validate metadata, by calcualating the hash (asset_id) and compare this to the
        given asset_id, if both are equal then the metadata is valid
        :param asset_id: asset id to compare with
        :param metadata: dict of metadata to calculate the hash ( asset_id)
        :return bool True if metadata is valid for the asset_id provided
        """
        if metadata_text:
            # the calc asset_id from the metadata should be same as this asset_id
            metadata_id = SurferModel.get_asset_id_from_metadata(metadata_text)
            if metadata_id != asset_id:
                logger.debug(f'metdata does not match {metadata_id} != {asset_id}')
            return metadata_id == asset_id
        return False


    @staticmethod
    def get_asset_id_from_metadata(metadata_text):
        """
        return the asset_id calculated from the metadata
        :param metadata: dict of metadata to hash
        a 64 char hex string, which is the asset id
        :return 64 char hex string, with no leading '0x'
        """
        return Web3.toHex(Web3.sha3(metadata_text.encode()))[2:]

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        SurferModel._http_client = http_client

    @staticmethod
    def get_authorization_token(surfer_url, surfer_username, surfer_password):
        """Get a surfer authorization token (create one if needed).
        Throws exception on error."""
        token_url = surfer_url + SURFER_BASE_URI + '/auth/token'
        r = requests.get(token_url, auth=(surfer_username, surfer_password))
        token = None
        if r.status_code == 200:
            tokens = r.json()
            if len(tokens) > 0:
                token = tokens[-1]
            else: # need to create a token
                r = requests.post(token_url, auth=(surfer_username, surfer_password))
                if r.status_code == 200:
                    token = r.json()
                else:
                    msg = f'unable to create token, status {r.status_code}'
                    logger.error(msg)
                    raise ValueError(msg)
        else:
            msg = f'unable to get tokens, status {r.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        logger.debug(f'using surfer token {token}')
        return token

    @staticmethod
    def find_supported_service(search_name_type):
        """ return the supported service record if the name or service type is found
        else return None """
        for name, service in SUPPORTED_SERVICES.items():
            if service['type'] == search_name_type:
                return service
            if name == search_name_type:
                return service
        return None

    @staticmethod
    def generate_service_endpoints(url):
        """
        Return a dict list of services available for this surfer
        in the format::

            {'name': service name, 'type': service_type, 'url': service endpoint url}

        """
        result = []
        for name, service in SUPPORTED_SERVICES.items():
            service_uri = service['uri']
            result.append({
                'name': name,
                'type': service['type'],
                'url': f'{url}{service_uri}',
            })
        return result



"""
    test_02_asset_identity


    As a developer working with Ocean,
    I need a stable identifier (Asset ID) for an arbitrary asset in the Ocean ecosystem

"""

import secrets
from starfish.asset import MemoryAsset
from starfish.agent import SurferAgent


def test_02_asset_register(ocean, resources, surfer_agent):
    test_data = secrets.token_hex(1024)
    asset1 = MemoryAsset(data=test_data)
    asset2 = MemoryAsset(data=test_data)
    listing = surfer_agent.register_asset(asset2, resources.listing_data)
    assert(not listing is None)
    assert(64==len(listing.listing_id))
    assert(listing.asset.data == asset2.data)
    assert(listing.asset.asset_id == asset2.asset_id)


def test_02_asset_upload(ocean, resources, surfer_agent):
    asset = MemoryAsset(data=secrets.token_hex(1024))
    listing = surfer_agent.register_asset(asset, resources.listing_data)
    surfer_agent.upload_asset(listing.asset)



"""
    Test ocean class

"""

import datetime
import pathlib
import json
import logging
import time
import secrets
import pytest

from web3 import Web3

from starfish import Ocean
from starfish.models.squid_model import SquidModel
from starfish.agent import SquidAgent
from starfish.asset import (
    FileAsset,
    RemoteAsset,
)
from starfish.exceptions import StarfishAssetNotFound

from squid_py.agreements.service_factory import ServiceDescriptor
from squid_py.utils.utilities import generate_new_id

from squid_py.brizo.brizo_provider import BrizoProvider
from squid_py.aquarius.aquarius import Aquarius
from squid_py.ddo.ddo import DDO

logger = logging.getLogger('test.core.test_squid')

def _register_asset_for_sale(agent, resources, account):

    asset = FileAsset(filename=resources.asset_file)
    listing = agent.register_asset(asset, resources.listing_data, account=account)
    assert(listing)
    assert(listing.asset.did)
    return(listing)


def test_asset_file_register(ocean, config, resources):
    publisher_account = ocean.get_account(config.publisher_account)

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    asset = FileAsset(filename=resources.asset_file)
    listing = agent.register_asset(asset, resources.listing_data, publisher_account)
    assert(listing)


def test_asset_remote_register(ocean, config, resources):
    publisher_account = ocean.get_account(config.publisher_account)

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)

    asset = RemoteAsset(url=resources.asset_remote)

    listing = agent.register_asset(asset, resources.listing_data, publisher_account)
    assert(listing)


def is_purchase_allowed(did, agreement_id, publish_account, conusmer_account):

    logger.debug(f'is purchase allowed {did}, {agreement_id}, {publish_account}, {conusmer_account}')
    return True

def test_asset(ocean, config, resources):

    agent = SquidAgent(ocean, config.squid_config)
    assert(agent)


    # test node has the account #0 unlocked
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()
    publisher_account.request_tokens(20)

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert(listing)
    assert(publisher_account)

    listing_did = listing.asset.did
    # start to test getting the asset from storage
    listing = agent.get_listing(listing_did)
    assert(listing)
    assert(listing.asset.did == listing_did)



    purchase_account = ocean.get_account(config.purchaser_account)
    logger.info(f'purchase_account {purchase_account.ocean_balance}')

    # test is_purchased for an account only
    assert(not listing.is_purchased(purchase_account))

    purchase_account.unlock()

    purchase_account.request_tokens(10)


    time.sleep(1)
    logger.info(f'purchase_account after token request {purchase_account.ocean_balance}')


    agent.start_agreement_events_monitor(publisher_account, is_purchase_allowed)

    # test purchase an asset
    purchase_asset = listing.purchase(purchase_account)
    assert(purchase_asset)

    assert(not purchase_asset.is_completed)

    error_message = purchase_asset.wait_for_completion()
    assert(error_message == True)

    assert(purchase_asset.is_completed)

    # assert Web3.toHex(event.args['_agreementId']) == agreement_id
    # assert len(os.listdir(consumer_ocean_instance.config.downloads_path)) == downloads_path_elements + 1

    # This test does not work with the current barge

    assert(purchase_asset.is_purchased)
    assert(purchase_asset.is_purchase_valid)


    # test getting a list of purchase ids from a listing
    purchase_ids = listing.get_purchase_ids
    assert(purchase_ids)
    assert(len(purchase_ids) == 1)
    assert(purchase_ids[0] == purchase_asset.purchase_id)

    # test is_purchased for an account only
    assert(listing.is_purchased(purchase_account))

    remote_asset = purchase_asset.consume_asset
    assert(remote_asset)


def test_search_listing(ocean, config, resources):


    agent = SquidAgent(ocean, config.squid_config)

    # test node has the account #0 unlocked
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    # choose a word from the description field
    text = resources.listing_data['author']
    words = text.split(' ')
    word = words[0]

    # should return at least 1 or more assets
    logger.info(f'search word is {word}')
    search_result = agent.search_listings(word)
    assert(search_result)

    assert(len(search_result) > 1)

def test_get_listing(ocean, config, resources):
    agent = SquidAgent(ocean, config.squid_config)
    # test node has the account #0 unlocked
    publisher_account = ocean.get_account(config.publisher_account)
    publisher_account.unlock()

    listing = _register_asset_for_sale(agent, resources, publisher_account)
    assert listing
    assert publisher_account

    found_listing = agent.get_listing(listing.listing_id)
    assert(found_listing)

    # now test for an asset in aquarius but not on the block chain network

    aquarius = Aquarius(config.squid_config['aquarius_url'])
    ddo = aquarius.get_asset_ddo(listing.listing_id)
    dummy_listing_id = f'did:op:{generate_new_id()}'
    ddo_dict = ddo.as_dictionary()
    ddo_dict['id'] = dummy_listing_id
    # fix squid index bug in service list ?
    index = 0
    for service in ddo_dict['service']:
        ddo_dict['service'][index]['serviceDefinitionId'] = index

    dummy_ddo = DDO(dictionary=ddo_dict)
    aquarius.publish_asset_ddo(dummy_ddo)

    ddo = aquarius.get_asset_ddo(dummy_listing_id)
    assert(ddo.did == dummy_listing_id)

    # the final test, the asset did is in aquarius, but the did -> url is not on the block chain
    with pytest.raises(StarfishAssetNotFound):
        dummy_listing = agent.get_listing(dummy_listing_id)

    # now try to purchase an invalid asset with no block chain DID
    purchase_account = ocean.get_account(config.purchaser_account)

    listing_list = agent.search_listings({'tags': ['starfish']})
    assert(listing_list)
    for listing in listing_list:
        assert(listing)
        if listing.listing_id == dummy_listing_id:
            with pytest.raises(StarfishAssetNotFound):
                listing.purchase(purchase_account)



"""
    test_squid_asset

"""

import secrets
import logging
import json

from starfish.asset import RemoteAsset


def test_init(metadata):
    asset = RemoteAsset(metadata)
    assert(asset)
    assert(isinstance(asset, RemoteAsset))




"""
    test_asset

"""

import secrets
import logging
import json

from starfish.asset import Asset

ASSET_METADATA = {
    'name': 'Asset',
    'type': 'asset',
}


def test_init(metadata):
    asset = Asset(ASSET_METADATA)
    assert(asset)
    assert(isinstance(asset, Asset))





from unittest.mock import Mock
import pytest
import secrets
import logging
import pathlib

from tests.integration.libs.integration_test_config import IntegrationTestConfig


from starfish import Ocean
from starfish.agent import (
    SurferAgent,
    SquidAgent,
)

from starfish.models.surfer_model import SurferModel

INTEGRATION_PATH = pathlib.Path.cwd() / 'tests' / 'integration'
CONFIG_FILE_PATH = INTEGRATION_PATH / 'config.ini'

@pytest.fixture(scope="module")
def ocean():
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    ocean = Ocean(
        keeper_url=integrationTestConfig.keeper_url,
        contracts_path=integrationTestConfig.contracts_path,
        gas_limit=integrationTestConfig.gas_limit,
        log_level=logging.WARNING
    )

    return ocean

@pytest.fixture(scope="module")
def config():
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    return integrationTestConfig

@pytest.fixture(scope="module")
def surfer_agent(ocean):
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)

    ddo_options = None
    if integrationTestConfig.koi_url:
        ddo_options = {
            'invoke': f'{integrationTestConfig.koi_url}/api/v1',
        }
    ddo = SurferAgent.generate_ddo(integrationTestConfig.surfer_url, ddo_options)
    options = {
        'url': integrationTestConfig.surfer_url,
        'username': integrationTestConfig.surfer_username,
        'password': integrationTestConfig.surfer_password,
    }
    return SurferAgent(ocean, did=ddo.did, ddo=ddo, options=options)

@pytest.fixture(scope="module")
def squid_agent(ocean):
    integrationTestConfig = IntegrationTestConfig(CONFIG_FILE_PATH)
    return SquidAgent(ocean, integrationTestConfig.squid_config)



import os
import logging
import json
import time

from unittest.mock import Mock
from eth_utils import add_0x_prefix

from squid_py import ConfigProvider
from squid_py.brizo.brizo import Brizo
from squid_py.did import id_to_did, did_to_id
from squid_py.agreements.service_agreement import ServiceAgreement
from squid_py.agreements.service_types import ServiceTypes
from squid_py.keeper.web3_provider import Web3Provider
from squid_py.keeper import Keeper
from squid_py.keeper.events_manager import EventsManager
# from squid_py.agreements.manager import AgreementsManager

logging.getLogger('urllib3').setLevel(logging.WARNING)
logging.getLogger('web3').setLevel(logging.WARNING)

logger = logging.getLogger('test.mocks.brizo_mock')

class BrizoMock(object):

    def __init__(self, ocean_instance=None, account=None):
        self._ocean_instance = ocean_instance
        self._is_event_subscribed = False
        self._ddo_records = {}
    """
    def subscribe(self, ocean, account, did, ddo):
        self._account = account
        model = ocean.get_squid_model()
        self._ocean_instance = model.get_squid_ocean(account)
        self._is_event_subscribed = False

        self._ddo_records[did] = ddo
        events_manager = EventsManager.get_instance(Keeper.get_instance())
        events_manager.stop_all_listeners()
        time.sleep(1)
        events_manager.agreement_listener._event_filters = dict()
        self._ocean_instance.agreements.subscribe_events(
            self._account.address,
            self._handle_agreement_created,
        )
        # at the moment we need to do this sleep or the event handle below
        # is not called. Not sure why?
        time.sleep(1)

    def _handle_agreement_created(self, event, *_):
#        print('_handle_agreement_created ', event)
        try:
            if not event or not event.args:
                print('no handle created')
                return

            self._is_event_subscribed = True
            print(f'Start handle_agreement_created: event_args={event.args}')
            config = ConfigProvider.get_config()
            provider_account = self._account
            assert provider_account.address == event.args['_accessProvider']

            did = id_to_did(event.args['_did'])
            agreement_id = Web3Provider.get_web3().toHex(event.args['_agreementId'])

            ddo = self._ddo_records[did]
            # calling the resolve calls the squid_py.aquarius module.
            # The module calls a http request session object created outside of this thread
            # and so can cause crashes
            #ddo = ocean.assets.resolve(did)
            sa = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS, ddo)

            condition_ids = sa.generate_agreement_condition_ids(
                agreement_id=agreement_id,
                asset_id=add_0x_prefix(did_to_id(did)),
                consumer_address=event.args['_accessConsumer'],
                publisher_address=ddo.publisher,
                keeper=Keeper.get_instance())

            events_manager = EventsManager.get_instance(Keeper.get_instance())
            agreements_manager = AgreementsManager(
                config,
                Keeper.get_instance(),
                events_manager,
                config.storage_path
            )
            agreements_manager.register_publisher(
                event.args['_accessConsumer'],
                agreement_id,
                did,
                sa,
                sa.service_definition_id,
                sa.get_price(),
                provider_account,
                condition_ids
            )
            print(f'handle_agreement_created() -- done registering event listeners.')
        except e as Exception:
            message = f'error with exception {e}'
            print(message)
            logger.error(message)
            raise

    @property
    def is_event_subscribed(self):
        return self._is_event_subscribed

    """

    def initialize_service_agreement(did, agreement_id, service_definition_id,
                                     signature, account_address, purchase_endpoint):
        print(f'BrizoMock.initialize_service_agreement: purchase_endpoint={purchase_endpoint}')
        self._ocean_instance.agreements.create(
            did,
            service_definition_id,
            agreement_id,
            signature,
            account_address,
            self._account
        )
        return True

    @staticmethod
    def consume_service(service_agreement_id, service_endpoint, account_address, files,
                        destination_folder, *_, **__):
        for f in files:
            with open(os.path.join(destination_folder, os.path.basename(f['url'])), 'w') as of:
                of.write(f'mock data {service_agreement_id}.{service_endpoint}.{account_address}')

    @staticmethod
    def invoke_service(service_agreement_id, service_endpoint, account_address, invoke_payload):
        """
        Call the Koi endpoint to get access to the different files that form the asset.

        :param service_agreement_id: Service Agreement Id, str
        :param service_endpoint: Url to consume, str
        :param account_address: ethereum address of the consumer signing this agreement, hex-str
        :param files: List containing the files to be consumed, list
        :param destination_folder: Path, str
        :return:
        """
        logger.info(f'invoke endpoint with this url: {service_endpoint}')

        return invoke_payload['params']

    @staticmethod
    def get_brizo_url(config):
        return Brizo.get_brizo_url(config)


    @staticmethod
    def get_purchase_endpoint(config):
        return f'{Brizo.get_brizo_url(config)}/services/access/initialize'

    @staticmethod
    def get_service_endpoint(config):
        return f'{Brizo.get_brizo_url(config)}/services/consume'



"""
    Memory Asset
"""

from starfish.asset.asset_base import AssetBase
from mimetypes import MimeTypes
from urllib.parse import urlparse


class RemoteAsset(AssetBase):
    """

    Remote asset can be used manage a remote asset using an URL on the ocean network

    :param metadata: Optional dictionary metadata to provide for the asset
        if non used then the class will generate a default metadata based on the data provided
    :type metadata: None or dict
    :param did: Optional did of the asset if it's registered
    :type did: None or str
    :param str filename: filename of the asset to register

    """
    def __init__(self, metadata=None, did=None, url=None):
        default_metadata = {
            'name': 'RemoteAsset',
            'type': 'remote',
            'contentType': 'application/octet-stream',
            'url': url,
        }
        if metadata is None:
            metadata = default_metadata
        if not isinstance(metadata, dict):
            raise ValueError('metadata must be a dict')
        metadata = AssetBase.merge_metadata(metadata, default_metadata)

        AssetBase.__init__(self, 'remote', metadata, did)
        self._url = metadata.get('url', url)

        if self._url and urlparse(self._url):
            mime = MimeTypes()
            mime_type = mime.guess_type(self._url)
            if mime_type and mime_type[0]:
                self._metadata['contentType'] = mime_type[0]

    @property
    def url(self):
        """

        Return the asset data

        :return: the asset data
        :type: str or byte
        """
        return self._url
